/*-
 * Copyright (c) 2010  Marin Atanasov Nikolov  <dnaeon@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/queue.h>
#include <sys/param.h>

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "main.h"
#include "config.h"

/*
 * This is the global config_head element,
 * that is used to access the configuration options
 */
STAILQ_HEAD(, config) config_head;

/* config_init
 *
 * initializes the configuration data
 */
void
config_init()
{

   STAILQ_INIT(&config_head);

   /* 
    * config_register_option(option, required, is_bool)
    */

   /* register the string options */
   config_register_option(CONFIG_PKG_ROOT, 1, 0);
   config_register_option(CONFIG_PKG_ARCH, 1, 0);
   config_register_option(CONFIG_PKG_RELEASE, 1, 0);
   config_register_option(CONFIG_PKG_URL, 1, 0);
   config_register_option(CONFIG_INDEX_FILE, 1, 0);
   config_register_option(CONFIG_FETCH_CMD, 1, 0);
   config_register_option(CONFIG_FETCH_CMD_ARGS, 1, 0);
   config_register_option(CONFIG_PKG_INSTALL_CMD, 1, 0);
   config_register_option(CONFIG_PKG_INSTALL_CMD_ARGS, 0, 0);
   config_register_option(CONFIG_PKG_INSTALL_CMD_REMOTE, 1, 0);
   config_register_option(CONFIG_PKG_INSTALL_CMD_REMOTE_ARGS, 0, 0);
   config_register_option(CONFIG_PKG_INFO_CMD, 1, 0);
   config_register_option(CONFIG_PKG_INFO_CMD_ARGS, 0, 0); 

   /* register the bool options */
   config_register_option(CONFIG_DEPS_FULL_TREE, 1, 1);
   config_register_option(CONFIG_SHOW_PORT_DIR, 1, 1);
   config_register_option(CONFIG_SHOW_INST_PREFIX, 1, 1);
   config_register_option(CONFIG_SHOW_DESC, 1, 1);
   config_register_option(CONFIG_SHOW_DESC_FILE, 1, 1);
   config_register_option(CONFIG_SHOW_MAINT, 1, 1);
   config_register_option(CONFIG_SHOW_CAT, 1, 1);
   config_register_option(CONFIG_SHOW_BDEPS, 1, 1);
   config_register_option(CONFIG_SHOW_RDEPS, 1, 1);
   config_register_option(CONFIG_SHOW_WWW, 1, 1);
   config_register_option(CONFIG_SHOW_EDEPS, 1, 1);
   config_register_option(CONFIG_SHOW_PDEPS, 1, 1);
   config_register_option(CONFIG_SHOW_FDEPS, 1, 1);
}

/* config_finish
 *
 * deallocates the memory used by the configuration data
 */
void
config_finish(void)
{
   CONFIG *entry;

   while (!STAILQ_EMPTY(&config_head)) {
      entry = STAILQ_FIRST(&config_head);
      free(entry->option);
      free(entry->value);
      STAILQ_REMOVE_HEAD(&config_head, entries);
      free(entry);
   }
}

/* config_check_options
 *
 * performs a check for valid configuration options.
 *
 * Returns 0 if configuration is OK, 1 otherwise
 */
int
config_check_options(void)
{
   CONFIG *entry;
   char flag;

   flag = 0;

   STAILQ_FOREACH(entry, &config_head, entries) {
      if (entry->required) {
	 if (entry->value == NULL) {
	    flag = 1;
	    warnx("Warning: Required option '%s' is not defined in the configuration file.", entry->option);
	 } else if (!STRNLEN(entry->value, MAXPATHLEN)) {
	    flag = 1;
	    warnx("Warning: Required option '%s' at line %d has no value.", entry->option, entry->line);
	 } 
	 else {
	    if (entry->is_bool && (config_get_bool_option(entry->option) == -1)) {
	       flag = 1;
	       warnx("Warning: Boolean option '%s' at line %d needs to be set to 'true' or 'false' only.", 
		  entry->option, entry->line);
	    }
	 }
      }
   }

   return(flag);
}

/* config_register_option
 *
 * registers a configuration option as a valid one.
 */
void
config_register_option(const char *option, char required, char is_bool)
{
   CONFIG *entry;

   entry = (CONFIG *) malloc(sizeof(CONFIG));
   entry->option = STRNDUP(option, MAXPATHLEN);
   entry->value  = NULL;

   if ((entry == NULL) || (entry->option == NULL)) {
      fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
      ERRX;
   }
   
   entry->line	   = 0;
   entry->required = required;
   entry->is_bool  = is_bool;

   STAILQ_INSERT_TAIL(&config_head, entry, entries);
}

/* config_read_file
 *
 * reads the specified configuration file and perfoms checks for properly defined syntax.
 */
void
config_read_file(const char *filename)
{
   FILE *fp;
   char buf[MAXPATHLEN];
   char *option, *value, *tmp;
   unsigned int line;

   if ((fp = fopen(filename, "r")) == NULL) { 
      fprintf(stderr, "%s: Cannot open configuration file.\n", PROGNAME);
      ERRX;
   }

   line = 0;

   while (fgets(buf, MAXPATHLEN - 1, fp)) {
      line++;

      tmp = buf;
      
      if (buf[0] == CONFIG_COMMENT || buf[0] == CONFIG_NEWLINE)
	 continue;

      option = strsep(&tmp, CONFIG_DELIMETER);
      value  = strsep(&tmp, CONFIG_DELIMETER);

      if ((option == NULL) || (value == NULL)) {
	 warnx("Error in configuration file at line %d.", line);
	 errx(1, "Option name and/or value missing at line %d.", line);
      }

      option = trim_spaces(option);
      value  = trim_spaces(value);

      /* 
       * if there are more values for the option in the configuration file
       * than the needed only one - throw an error
       */
      if (tmp) {
	 warnx("Error in configuration file at line %d.", line);
	 errx(1, "Too many values for option '%s'", option);
      }

      /* set the value for the configuration option */
      config_set_option(option, value, line);
   }

   fclose(fp);

   /* check if all required options have values defined */
   if (config_check_options())
      errx(1, "Error parsing configuration file.");

   /* expand options, that might contain other option definitions */
   if (config_expand_options())
      errx(1, "Error parsing configuration file."); 
}


/* config_expand_options
 *
 * expands any configuration options in the configuration file,
 * in the form of variables like $CONFIG_OPTION$.
 *
 * Returns 0 on success and 1 when there was an error during parsing of an option
 */
int
config_expand_options(void)
{
   struct config_expands {
      char *option;
      char *option_expanded;

      STAILQ_ENTRY(config_expands) entries;
   } *config_expand_entry;

   CONFIG *config_entry;

   char	  *buf, *tmp_option,
	  *tmp_value, *tmp_oldvalue, *tofree;
   int     flag, count;

   STAILQ_HEAD(, config_expands) config_expands_head;

   flag = count = 0;
  
   /* 
    * iterate through the configuration options in order
    * to find possible option expansions
    */ 
   STAILQ_FOREACH(config_entry, &config_head, entries) {

      STAILQ_INIT (&config_expands_head);
      tofree = tmp_value = STRNDUP(config_entry->value, MAXPATHLEN);

      while (*tmp_value) {

	 /* find the first CONFIG_EXPAND_DELIMITER character */   
	 if (*tmp_value++ == CONFIG_EXPAND_DELIMITER) {
	    count = 0;
	    buf = (char *) malloc(STRNLEN(config_entry->value, MAXPATHLEN) + 1);

	    if (buf == NULL) {
	       fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	       ERRX;
	    }

	    while (*tmp_value && *tmp_value != CONFIG_EXPAND_DELIMITER) {
	       count++;
	       *buf++ = *tmp_value++;
	    }

	    /* there was a missing CONFIG_EXPAND_DELIMITER */
	    if (*tmp_value != CONFIG_EXPAND_DELIMITER)
	       flag = 1;

	    else {
	       *buf = '\0';

	       /* locate the buffer at the beginning of the string */
	       buf -= count;
	       
	       config_expand_entry = (struct config_expands *) malloc(sizeof(struct config_expands));
	       config_expand_entry->option = STRNDUP(buf, MAXPATHLEN);

	       if ((config_expand_entry == NULL) || (config_expand_entry->option == NULL)) {
		  fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
		  ERRX;
	       }

	       STAILQ_INSERT_TAIL(&config_expands_head, config_expand_entry, entries);

	       tmp_value++;
	    }

	    free(buf);
	 }
      }

      /* if there were options to be expanded, check if they are valid ones */
      if (!STAILQ_EMPTY(&config_expands_head)) {
	 STAILQ_FOREACH(config_expand_entry, &config_expands_head, entries) {
	    config_expand_entry->option_expanded = config_get_option(config_expand_entry->option);
	    if (config_expand_entry->option_expanded == NULL) {
	       flag = 1;
	       warnx("Warning: Invalid option expansion at line %d in configuration file.", 
			config_entry->line);
	       warnx("Warning: Unknown configuration option '%s' at line %d was found during expansion.",
			config_expand_entry->option, config_entry->line);
	    }
	 }

	 /* if we have something to expand */
	 if (!flag) {
	    STAILQ_FOREACH(config_expand_entry, &config_expands_head, entries) {
	       /* 
		* tmp_option = (char *) malloc(STRNLEN(tmp->option + 3, MAXPATHLEN)
		* 
		* 1 byte for NULL terminator
		* 2 bytes for the CONFIG_EXPAND_DELIMITER characters
		*/
	       tmp_option = (char *) malloc(STRNLEN(config_expand_entry->option, MAXPATHLEN) + 3);
	       snprintf (tmp_option, STRNLEN(config_expand_entry->option, MAXPATHLEN) + 3, "$%s$", config_expand_entry->option);

	       tmp_oldvalue = config_entry->value;

	       config_entry->value = str_replace(tmp_oldvalue, tmp_option, config_expand_entry->option_expanded);

	       free(tmp_option);
	       free(tmp_oldvalue);
	    }
	 }
	       
	 /* destroy the queue for the next loop */
	 while (!STAILQ_EMPTY(&config_expands_head)) {
	    config_expand_entry = STAILQ_FIRST(&config_expands_head);
	    free(config_expand_entry->option);
	    STAILQ_REMOVE_HEAD(&config_expands_head, entries);
	    free(config_expand_entry);
	 }
      }
      free(tofree);
   }

   return(flag);
}

/* config_set_option
 *
 * sets the needed data to identify an option. Also does some checks for options,
 * that are being redefined in the configuration file.
 */
void
config_set_option(const char *option, const char *value, unsigned int line)
{
   CONFIG *entry;

   char flag = 0;

   STAILQ_FOREACH(entry, &config_head, entries) {
      if (!strncmp(entry->option, option, MAXPATHLEN)) {
	 /* if the option has already been defined */
	 if (entry->value) {
	    warnx("Error in configuration file at line %d.", line);
	    errx(1, "Option '%s' is already defined at line %d.", entry->option, entry->line);
	 }

	 flag = 1;
	 entry->value = STRNDUP(value, MAXPATHLEN);
	 entry->line  = line;
      }
   }

   /* if flag == 0, there is a non-registered option */
   if (!flag) {
      warnx("Error in configuration file at line %d.", line);
      errx(1, "Unknown configuration option '%s'.", option);
   }
}

/* config_set_value
 *
 * sets/updates a value of an option from the configuration file
 */
void
config_set_value(const char *option, const char *newvalue)
{
   CONFIG *entry;

   STAILQ_FOREACH(entry, &config_head, entries) 
      if (!strncmp(entry->option, option, MAXPATHLEN)) {
	 /* free the old value first */
	 if (entry->value != NULL)
	    free(entry->value);
	 entry->value = STRNDUP(newvalue, MAXPATHLEN);
      }
}

/* config_get_option
 *
 * returns the value of an option, and NULL if there is 
 * no such an option registered.
 */
char *
config_get_option(const char *option)
{
   CONFIG *entry;

   STAILQ_FOREACH(entry, &config_head, entries)
      if (!strncmp(entry->option, option, MAXPATHLEN))
	 return(entry->value);

   return(NULL);
}

/* get_bool_option
 *
 * used to check whether a bool option is set to 'true' or 'false'
 *
 * Returns 1 for options defined to 'true' and 0 for false.
 */
int
config_get_bool_option(const char *option)
{
   CONFIG *entry;

   char *t = "true";
   char *f = "false";

   STAILQ_FOREACH(entry, &config_head, entries) {
      if (!strncmp(entry->option, option, MAXPATHLEN)) { 
	 if (!strncmp(entry->value, t, MAXPATHLEN))
	    return(1);
	 if (!strncmp(entry->value, f, MAXPATHLEN))
	    return(0);
      }
   }

   return(-1);
}

/* str_replace
 *
 * replaces a string with a new one.
 * This function is mainly used by the config_option_expand() function
 *
 * Returns the newly replaced string
 */
char *
str_replace(char *str, char *old, char *new)
{
   char *ret;
   int   i, count, 
	 newlen, oldlen;
 
   count  = 0;
   newlen = STRNLEN(new, MAXPATHLEN);
   oldlen = STRNLEN(old, MAXPATHLEN);

   for (i = 0; str[i]; i++)
      if (strstr(&str[i], old) == &str[i]) {
	 count++;
	 i += oldlen - 1;
      }
 
   ret = (char *) malloc(i + 1 + (count * (newlen - oldlen)));
  
   if (ret == NULL) {
      fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
      ERRX;
   }

   i = 0;

   while (*str) {
      if (strstr(str, old) == str) {
	 strcpy(&ret[i], new);
	 i += newlen;
	 str += oldlen;
      } else
	 ret[i++] = *str++;
   }
   
   ret[i] = '\0';

   return(ret);
}   
      
/* trim_spaces
 *
 * trims any leading and/or tailing spaces from a string
 *
 * Returns the trimmed string
 */
char *
trim_spaces(char *str)
{
   char *buf;
   int   i, len;

   buf = str;
   len = STRNLEN(str, MAXPATHLEN);
   
   for (i = 0; i < len && isspace(str[i]); i++, buf++) ;
   for (i = len - 1; i >= 0 && isspace(str[i]); str[i] = '\0', i--) ;

   return(buf);
}

