/*-
 * Copyright (c) 2010  Marin Atanasov Nikolov  <dnaeon@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/param.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <signal.h>

#include "pkg.h"
#include "main.h"
#include "config.h"
#include "scan_dir.h"
#include "scan_index.h"

static void usage(void);

/* main
 *
 * pkg_add_it - a tool for interactive installation of packages for FreeBSD
 */
int
main (int argc, char *argv[])
{
   /* options struct */
   static struct option longopts[] = {
      { "help", no_argument, NULL, 'h' },
      { "dry-run", no_argument, NULL, 'n' },
      { "pager", no_argument, NULL, 'p' },
      { "show-cats", no_argument, NULL, 'C' },
      { "regex", no_argument, NULL, 'x' },
      { "extended", no_argument, NULL, 'X' },
      { "index", required_argument, NULL, 'i' },
      { "file", required_argument, NULL, 'f' },
      { "local", required_argument, NULL, 'l' },
      { "remote", required_argument, NULL, 'r' },
      { "show-deps", required_argument, NULL, 'D' },
      { NULL, 0, NULL, 0 }
   } ;
        
   int chopt, numfiles, tmp_optind,
       file_optind, index_optind, cmd_line_opts;

   tmp_optind = file_optind = index_optind = cmd_line_opts = 0;
   
   /* make stdout unbuffered */
   setvbuf(stdout, (char *)NULL, _IONBF, 0);

   /* 
    * ignore SIGPIPE.
    *
    * when using the --pager option we might receive a 
    * SIGPIPE signal that stops the program from being
    * executed further, so in order to avoid this,
    * we just ingore the signal.
    */
   signal(SIGPIPE, SIG_IGN);

    /* initialize the configuration and package data */
   config_init();
   pkg_init();

   cmd_line_opts = 0;

   /* check for valid command-line arguments */
   while ((chopt = getopt_long(argc, argv, CL_ARGS, longopts, NULL)) != -1) {
      switch(chopt) {
      case 'l': 
	    cmd_line_opts |= OPTION_LOCAL_INSTALL;
	    cmd_line_opts |= OPTION_VALID_OPTS;
	    tmp_optind	   = optind;
            break;
      case 'r':
	    cmd_line_opts |= OPTION_REMOTE_INSTALL;
	    cmd_line_opts |= OPTION_VALID_OPTS;
	    tmp_optind	   = optind;
	    break;
      case 'D':
	    cmd_line_opts |= OPTION_SHOW_DEPS;
	    cmd_line_opts |= OPTION_REMOTE_INSTALL;
	    cmd_line_opts |= OPTION_VALID_OPTS;
	    tmp_optind	   = optind;
	    break;
      case 'C':
	    cmd_line_opts |= OPTION_SHOW_CATS;
	    cmd_line_opts |= OPTION_VALID_OPTS;
	    break;
      case 'x':
	    cmd_line_opts |= OPTION_REGEX;
	    break;
      case 'X':
	    cmd_line_opts |= OPTION_REGEX;
	    cmd_line_opts |= OPTION_REGEX_EXTENDED;
	    break;
      case 'n':
	    cmd_line_opts |= OPTION_DRY_RUN;
            break;
      case 'p':
	    cmd_line_opts |= OPTION_PAGER;
	    break;
      case 'f':
	    cmd_line_opts |= OPTION_ALTERNATIVE_FILE;
	    file_optind	   = optind - 1;
	    break;
      case 'i':
	    cmd_line_opts |= OPTION_ALTERNATIVE_INDEX;
	    index_optind   = optind - 1;
	    break;
      case 'h':
	    /* FALLTHROUGH */
      default:
            usage(); 
	    /* NOTREACHED */
      }
   } 

   /* 
    * Check if there were any valid command-line options given
    * Note: a valid command-line option is considered to be either a
    * local or remote install mode.
    */
   if (!(cmd_line_opts & OPTION_VALID_OPTS)) {
      fprintf(stderr, "==>> Invalid call to %s\n", PROGNAME);
      fprintf(stderr, "==>> Run %s --help for a list of available options\n", PROGNAME);
      exit(1);
   } 

   /* Actions that are allowed to be performed are:
    *
    *	 - local install
    *	 - remote install
    *	 - show dependencies
    *	 - show categories
    *
    * Combining together any of the above actions is considered ambiguous
    */
   if (((cmd_line_opts & OPTION_LOCAL_INSTALL)  && \
       (cmd_line_opts & OPTION_REMOTE_INSTALL)) || \
       ((cmd_line_opts & OPTION_LOCAL_INSTALL)  && \
       (cmd_line_opts & OPTION_SHOW_CATS))      || \
       ((cmd_line_opts & OPTION_REMOTE_INSTALL) && \
       (cmd_line_opts & OPTION_SHOW_CATS)))	   {
      fprintf(stderr, "==>> Ambiguous call to %s\n", PROGNAME);
      fprintf(stderr, "==>> Run %s --help for a list of available options\n", PROGNAME);
      exit(1);
   }

   /* read configuration file */
   (cmd_line_opts & OPTION_ALTERNATIVE_FILE) ? config_read_file(argv[file_optind])
					     : config_read_file(CONFIG_DEFAULT_FILE);

   /* set new index file, if specified from command-line */
   if (cmd_line_opts & OPTION_ALTERNATIVE_INDEX)
      config_set_value(CONFIG_INDEX_FILE, argv[index_optind]);

   if (cmd_line_opts & OPTION_SHOW_CATS) {
      numfiles = pkg_find_cats();
      pkg_display_cats(numfiles);
   } else {
      numfiles = (cmd_line_opts & OPTION_LOCAL_INSTALL) ? pkg_local_install(argc, argv, tmp_optind, cmd_line_opts) 
						        : pkg_remote_install(argc, argv, tmp_optind, cmd_line_opts);
      if (!numfiles) {
	 fprintf(stderr, "==>> No matching packages were found.\n");
	 fprintf(stderr, "==>> Perhaps a wrong package name was given?\n");
      } else {
	 pkg_display_found(numfiles, cmd_line_opts & OPTION_PAGER);
	      
	 /* install any packages? */
	 if (!(cmd_line_opts & OPTION_DRY_RUN)) {
	    pkg_choose(numfiles);
	    pkg_install_packages(cmd_line_opts);
	 } 
      }
   }

   config_finish();
   pkg_finish();
   
   return(0);
}

/* yesno_prompt
 *
 * Creates a YES/NO prompt that ask the user to make a choice.
 *
 * Returns the chosen decision.
 */
int
yesno_prompt(const char *prompt)
{
   int choice;

   do { 
      printf(prompt);
      choice = fgetc(stdin);
      fpurge(stdin);
   } while ((choice != 'y') && (choice != 'Y') && \
	    (choice != 'n') && (choice != 'N'));

   return(choice);
}

/* usage
 * 
 * prints the usage of the program
 */
static void
usage(void)
{

   fprintf(stderr, "Usage: %s [-hnpxXC] [-i INDEX] [-f config] [-l <pkg-dir> <pkg-name>]\n", PROGNAME);
   fprintf(stderr, "                  [-r <pkg-name> <pkg-cat>] [-D <pkg-name> <pkg-cat>]\n");
   fprintf(stderr, "       %s -h|--help\n", PROGNAME);
   fprintf(stderr, "       %s -C|--show-cats\n", PROGNAME);
   fprintf(stderr, "       %s [-npxX] [-f config] -l <pkg-dir> <pkg-name>\n", PROGNAME);
   fprintf(stderr, "       %s [-npxX] [-f config] [-i INDEX] -r <pkg-name> [<pkg-cat>]\n", PROGNAME);
   fprintf(stderr, "       %s [-pxX]  [-f config] [-i INDEX] -D <pkg-name> [<pkg-cat>]\n\n", PROGNAME);
   fprintf(stderr, "%s - %s, v%s\n", PROGNAME, PROGDESC, PROGVERS);
   fprintf(stderr, "%s v%s : built on %s at %s\n\n", PROGNAME, PROGVERS, __DATE__, __TIME__);
   fprintf(stderr, "Available options:\n\n");
   fprintf(stderr, " -h|--help          Displays this help message\n");
   fprintf(stderr, " -n|--dry-run       Dry run - no installation of packages is done\n");
   fprintf(stderr, " -p|--pager         Use a pager when displaying packages/dependencies\n");
   fprintf(stderr, " -x|--regex         Treat <pkg-name> as a regular expression\n");
   fprintf(stderr, " -X|--extended      Treat <pkg-name> as an extended regular expression\n");
   fprintf(stderr, " -C|--show-cats     Displays categories and number of packages in them\n");
   fprintf(stderr, " -i|--index         Specify an alternative INDEX file\n");
   fprintf(stderr, " -f|--file          Specify an alternative configuration file\n");
   fprintf(stderr, " -l|--local         Specifies a local pattern search\n");
   fprintf(stderr, " -r|--remote        Specifies a remote pattern search\n");
   fprintf(stderr, " -D|--show-deps     Display dependencies only (implies -n and -r)\n\n");
   fprintf(stderr, " <pkg-dir>          Specifies the directory containing package files (.tbz)\n");
   fprintf(stderr, "                    Used only with the `--local` option\n");
   fprintf(stderr, " <pkg-name>         Specifies the package name\n");
   fprintf(stderr, " <pkg-cat>          Specifies package category to be searched (optional)\n");
   fprintf(stderr, "\nFor example usage and more information read %s(1)\n", PROGNAME);

   exit(1);
}

