/*-
 * Copyright (c) 2010  Marin Atanasov Nikolov  <dnaeon@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/queue.h>

#include <stdio.h>
#include <regex.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <unistd.h>

#include "pkg.h"
#include "main.h"
#include "config.h"
#include "scan_dir.h"
#include "scan_index.h"

/* A tail that contains all the root packages,
 * that were selected for install.
 *
 * This tail gets expanded to a multi-level 
 * dependency tree when doing remote install and 
 * DEPS_FULL_TREE=true.
 */
static STAILQ_HEAD(pkg_head_type, pkg) pkg_head;

/* A tail that contains only one-level of
 * package dependencies.
 *
 * This tail contains all the packages that are 
 * about to be installed on the system.
 *
 * This tail is also used to build the full tree
 * of dependencies stored in 'pkg_head', since it
 * contains all the information about the packages.
 */
static STAILQ_HEAD(, pkg) pkg_single_head;

/* pkg_init
 *
 * Initializes the pkg_* tails 
 */
void
pkg_init(void)
{
   STAILQ_INIT(&pkg_head);
   STAILQ_INIT(&pkg_single_head);
}

/* pkg_insert
 *
 * Saves a given package to the pkg_head tail
 *
 * Here we only copy the data that we need.
 * 
 * In the configuration file the SHOW_* options define the fields
 * to be displayed. Here we only allocate memory for those fields
 * that are set to true in the configuration file.
 *
 * The only two elements that we always allocate memory for are the
 * pkg_fields[PKG_NAME] and pkg_fields[PKG_CAT] elements, since only
 * they are required to know, in order to install a package.
 *
 * When doing a remote install of packages we also need to allocate
 * memory for the pkg_fields[PKG_RDEPS] element in order to have 
 * the package run-dependencies.
 *
 * Arguments:
 *    
 *    entry - the package entry to be added to the tail
 *    rdeps - set to 1 when doing a remote package install,
 *	      and 0 when doing a local install.
 */
void
pkg_insert(PKG *entry, int rdeps)
{
   PKG *newentry;
   int  show[PKG_FIELDS], i;
   
   show[PKG_NAME]	   = 1; 
   show[PKG_PORT_DIR]	   = config_get_bool_option(CONFIG_SHOW_PORT_DIR);
   show[PKG_INST_PREFIX]   = config_get_bool_option(CONFIG_SHOW_INST_PREFIX);
   show[PKG_DESC]	   = config_get_bool_option(CONFIG_SHOW_DESC);
   show[PKG_DESC_FILE]     = config_get_bool_option(CONFIG_SHOW_DESC_FILE);
   show[PKG_MAINT]	   = config_get_bool_option(CONFIG_SHOW_MAINT);
   show[PKG_CAT]	   = 1;
   show[PKG_BDEPS]	   = config_get_bool_option(CONFIG_SHOW_BDEPS);
   show[PKG_RDEPS]	   = rdeps;
   show[PKG_WWW]	   = config_get_bool_option(CONFIG_SHOW_WWW);
   show[PKG_EDEPS]	   = config_get_bool_option(CONFIG_SHOW_EDEPS);
   show[PKG_PDEPS]	   = config_get_bool_option(CONFIG_SHOW_PDEPS);
   show[PKG_FDEPS]	   = config_get_bool_option(CONFIG_SHOW_FDEPS);

   newentry = (PKG *) malloc(sizeof(PKG));
   
   if (newentry == NULL) {
      fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
      ERRX;
   }

   newentry->flags = 0;
   newentry->pkg_num = 1;
   STAILQ_INIT(&newentry->deps_head);

   /* flag the package as root package */
   newentry->flags |= PKG_FLAGS_ROOT_PKG;

   for (i = 0; i < PKG_FIELDS; i++) 
      newentry->pkg_fields[i] = NULL;
   
   for (i = 0; i < PKG_FIELDS; i++)
      if ((entry->pkg_fields[i] != NULL) && show[i] && \
		      (*(entry)->pkg_fields[i] != '\0')) {
	    newentry->pkg_fields[i] = STRNDUP(entry->pkg_fields[i], \
					      INDEX_MAX_LINE_LEN);
	    
	    if (newentry->pkg_fields[i] == NULL) {
	       fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	       ERRX;
	    }
      }

   STAILQ_INSERT_TAIL(&pkg_head, newentry, entries);
}

/* pkg_choose
 * 
 * Prompts the user to select packages for installation.
 *
 * Arguments:
 *    
 *    - num_pkgs - the number of packages that are
 *		   matching the search pattern.
 */
void
pkg_choose(int num_pkgs)
{
   PKG	    *entry, *entry_safe;
   unsigned  packages[MAXPATHLEN];
   char	     buf[MAXPATHLEN], *tofree, *token, *str;
   int	     i, j, k, m, flag;

   printf ("==>> Please specify a list of package numbers, separated by commas.\n\n");

   do {
      i = 0;
      flag = 1;

      printf ("==>> Choose the package(s) you wish to install [1-%d]: ", num_pkgs);
      fgets(buf, MAXPATHLEN - 1, stdin);
      
      tofree = str = STRNDUP(buf, MAXPATHLEN);
      
      if (tofree == NULL) {
	 fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	 ERRX;
      }

      while ((token = strsep(&str, ",")) != NULL) {
	 if ((atoi(token) <= 0) || (atoi(token) > num_pkgs)) {
	    flag = 0;
	    break;
	 }
	 packages[i++] = atoi(token);
      }

      free(tofree);

      if (!flag) 
	 printf ("==>> Invalid input has been detected, please try again.\n\n");

   } while(!flag);

   /* make zeros any duplicate packages you might have selected */
   for (j = 0; j < i; j++) 
      for (k = j + 1; k < i; k++) 
	 if (packages[k] == packages[j])
	    packages[k] = 0;

   /* 
    * leave only the packages that are chosen to be installed,
    * and remove the rest of packages from the tail
    *
    * k    == current package index
    * flag == 0, remove package from tail
    * flag == 1, leave package in tail
    */

   k = 1;

   STAILQ_FOREACH_SAFE(entry, &pkg_head, entries, entry_safe) {
      flag = 0;
      for (j = 0; j < i; j++)
	 if ((packages[j] != 0) && (packages[j] == k))
	    flag = 1;

      /* remove the package from the tail */
      if (!flag) {
	 for (m = 0; m < PKG_FIELDS; m++)
	    if(entry->pkg_fields[m] != NULL)
	       free(entry->pkg_fields[m]);

	 STAILQ_REMOVE(&pkg_head, entry, pkg, entries);
	 free(entry);
      }
      k++;
   }
} 

/* pkg_display_cats
 *
 * Displays all primary categories from the INDEX files,
 * and the number of packages in each of them.
 *
 * Arguments:
 *    
 *    - cats   - the number of found categories
 */
void
pkg_display_cats(int cats_num)
{
   PKG	   *entry;
   unsigned i, total_pkgs;
  
   printf("==>> Total categories found: %u\n", cats_num);
   printf("==>> Displaying all primary categories:\n\n%s", LONGLINE);

   i = total_pkgs = 0;

   STAILQ_FOREACH(entry, &pkg_head, entries) {
      printf("\n(%.3d)\t%s%-15s\n", ++i, PKG_CAT_HEADER, \
	      entry->pkg_fields[PKG_CAT]);
      
      printf("\tContains: %u package(s)\n", entry->pkg_num);
      total_pkgs += entry->pkg_num;
   }
  
   printf("\n%s\n==>> Total categories in INDEX: %u\n", LONGLINE, cats_num);
   printf("==>> Total packages in INDEX: %u\n", total_pkgs);
}

/* pkg_display_found
 *
 * Displays all packages that were found during the 
 * initial package search.
 *
 * Arguments:
 *
 *    - num_pkgs - the number of packages that were found
 *    - pager	 - when true, use a pager, otherwise output to stdout
 */
void
pkg_display_found(int num_pkgs, int pager)
{
   PKG  *entry;
   FILE *pipe;
   char *buf;
   int   i, j;

   const char *pkg_fields_headers[] = { PKG_NAME_HEADER, PKG_PORT_DIR_HEADER,
					PKG_INST_PREFIX_HEADER, PKG_DESC_HEADER,
				        PKG_DESC_FILE_HEADER, PKG_MAINT_HEADER,
				        PKG_CAT_HEADER, PKG_BDEPS_HEADER,
				        PKG_RDEPS_HEADER, PKG_WWW_HEADER,
				        PKG_EDEPS_HEADER, PKG_PDEPS_HEADER,
				        PKG_FDEPS_HEADER
				      };
			        	
   int show[PKG_FIELDS];

   show[PKG_PORT_DIR]	   = config_get_bool_option(CONFIG_SHOW_PORT_DIR);
   show[PKG_INST_PREFIX]   = config_get_bool_option(CONFIG_SHOW_INST_PREFIX);
   show[PKG_DESC]	   = config_get_bool_option(CONFIG_SHOW_DESC);
   show[PKG_DESC_FILE]     = config_get_bool_option(CONFIG_SHOW_DESC_FILE);
   show[PKG_MAINT]	   = config_get_bool_option(CONFIG_SHOW_MAINT);
   show[PKG_CAT]	   = config_get_bool_option(CONFIG_SHOW_CAT);
   show[PKG_BDEPS]	   = config_get_bool_option(CONFIG_SHOW_BDEPS);
   show[PKG_RDEPS]	   = config_get_bool_option(CONFIG_SHOW_RDEPS);
   show[PKG_WWW]	   = config_get_bool_option(CONFIG_SHOW_WWW);
   show[PKG_EDEPS]	   = config_get_bool_option(CONFIG_SHOW_EDEPS);
   show[PKG_PDEPS]	   = config_get_bool_option(CONFIG_SHOW_PDEPS);
   show[PKG_FDEPS]	   = config_get_bool_option(CONFIG_SHOW_FDEPS);
 
   buf = getenv("PAGER");

   pipe = ((buf == NULL) || !pager) ? stdout : popen(buf, "w");
  
   if (pipe == NULL)
      pipe = stdout;

   fprintf(pipe, "==>> Total packages found: %d\n", num_pkgs);
   fprintf(pipe, "==>> List of packages matching the given pattern: \n\n%s", LONGLINE);

   i = 0;

   STAILQ_FOREACH(entry, &pkg_head, entries) {
      fprintf(pipe, "\n(%.3d)\t%s%-30s", ++i, pkg_fields_headers[PKG_NAME], \
	      entry->pkg_fields[PKG_NAME]);
      
      /* display package category */
      if (show[PKG_CAT])
	 fprintf(pipe, " (in %s)", entry->pkg_fields[PKG_CAT]);

      /* display the rest of the fields, if there are any */
      for (j = PKG_PORT_DIR; j < PKG_FIELDS; j++)
	 if (show[j] && (entry->pkg_fields[j] != NULL) && (j != PKG_CAT))
	     fprintf(pipe, "\n\t%s%s", pkg_fields_headers[j], entry->pkg_fields[j]);
      
      fprintf(pipe, "\n");
   }
  
   fprintf(pipe, "\n%s\n==>> Total packages found: %d\n", LONGLINE, num_pkgs);

   (pager) ? pclose(pipe) : 0;
}

/* pkg_display_deps
 *
 * Display the packages and their dependencies in a 
 * non-tree, one-level depth dependency view.
 *
 * Arguments:
 *    
 *    - pipe - print to a pipe, or stdout
 *
 * Returns the number of packages that are about to be installed,
 * and false if there are none.
 */
int
pkg_display_deps(FILE *pipe)
{
   PKG *entry;
   int  i, res;
   
   res = 0;

   STAILQ_FOREACH(entry, &pkg_single_head, entries) {

      res += !(entry->flags & PKG_FLAGS_INSTALLED);

      fprintf(pipe, "%s", (entry->flags & PKG_FLAGS_INSTALLED) ? \
			"[   INSTALLED   ]"                        : \
			"[ NOT INSTALLED ]");

      for (i = 0; i < PKG_DEPS_INDENT; i++)
	 fputc(PKG_DEPS_INDENT_CHAR, pipe);

      fprintf(pipe, "%-30s\n", entry->pkg_fields[PKG_NAME]);
   
   }

   return(res);
}

/*
 * pkg_display_deps_full_tree
 *
 * Displays the packages that are about to be installed,
 * and their dependencies in a full-tree, multi-level depth view.
 *
 * Arguments:
 *
 *    - pkg_entry - the package and it's dependencies to be displayed.
 *    - root_pkg  - set to 1, when this is a package from the main
 *		    pkg_head tail.
 *    - selected  - set to 1, when this package is already registered as
 *		    as an existing package/dependency.
 *    - is_last   - set to 1, when this package is the last one of a dependency tail
 * 
 * Returns true, if there is at least one package that is not
 * currently installed (without counting the selected packages), false otherwise.
 */
int
pkg_display_deps_full_tree(PKG *pkg_entry, FILE *pipe, int root_pkg,
			   int selected, int is_last)
{
   PKG *entry, *deps_entry;

   static char *ps = "%s/%-30s\n";
   static char *ds = "\\___ %s/%-30s\n";
   char        *tofree, *token, *bufp;
   static int   indent, res = 0;
   int	        i, tmp_is_last;

   if (root_pkg == 1) {
      indent  = PKG_DEPS_INDENT;
      res = 0;
   } else 
       indent += PKG_DEPS_INDENT;

   res += !(pkg_entry->flags & PKG_FLAGS_INSTALLED) && \
	  !(pkg_entry->flags & PKG_FLAGS_SELECTED)  && !selected;

   fprintf(pipe, "%s", (pkg_entry->flags & PKG_FLAGS_INSTALLED)   ?  \
			"[   INSTALLED   ]"                       :  \
		         ((pkg_entry->flags & PKG_FLAGS_SELECTED) || \
			   selected)				  ?  \
			"[  PRESELECTED  ]"		          :  \
			"[ NOT INSTALLED ]");

   for (i = 0; i < indent; i++) {
      if (((STAILQ_NEXT(pkg_entry, entries) == NULL) && is_last) && \
	  (i == (indent - PKG_DEPS_INDENT)))
	 fputc(PKG_DEPS_INDENT_CHAR, pipe);
      else if ((STAILQ_NEXT(pkg_entry, entries) == NULL) && (i == (indent - PKG_DEPS_INDENT)))
	 fputc(PKG_DEPS_INDENT_LINK, pipe);
      else if (!(i % PKG_DEPS_INDENT) && ((i + 1) >= (PKG_DEPS_INDENT * 2)))
	 fputc(PKG_DEPS_INDENT_LINK, pipe);
      else
	 fputc(PKG_DEPS_INDENT_CHAR, pipe);
   }

   fputc((!root_pkg && (STAILQ_NEXT(pkg_entry, entries) != NULL) ? PKG_DEPS_INDENT_LINK 
							         : PKG_DEPS_INDENT_CHAR), pipe);

   tofree = bufp = STRNDUP(pkg_entry->pkg_fields[PKG_CAT], INDEX_MAX_LINE_LEN);
   
   if (tofree == NULL) { 
      fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
      ERRX;
   }

   token = strsep(&bufp, " ");

   fprintf(pipe, (root_pkg == 1) ? ps : ds, token, pkg_entry->pkg_fields[PKG_NAME]);

   free(tofree);

   indent += PKG_DEPS_INDENT;

   STAILQ_FOREACH(entry, &pkg_entry->deps_head, entries) {
      res += !(entry->flags & PKG_FLAGS_INSTALLED) && \
	     !(entry->flags & PKG_FLAGS_SELECTED)  && \
	     !(pkg_entry->flags & PKG_FLAGS_SELECTED) && !selected;
	    
      fprintf(pipe, "%s", (entry->flags & PKG_FLAGS_INSTALLED)    ?     \
			"[   INSTALLED   ]"                       :     \
		         ((entry->flags & PKG_FLAGS_SELECTED)	  ||    \
			  (pkg_entry->flags & PKG_FLAGS_SELECTED) ||    \
			  selected)				  ?     \
			"[  PRESELECTED  ]"		          :     \
			"[ NOT INSTALLED ]");

      for (i = 0; i < indent; i++) {
	 if ((STAILQ_NEXT(entry, entries) == NULL) && (i == (indent - PKG_DEPS_INDENT)) && \
	     (STAILQ_NEXT(pkg_entry, entries) == NULL))
	    fputc(PKG_DEPS_INDENT_CHAR, pipe);
	 else if (!(i % PKG_DEPS_INDENT) && ((i + 1) >= (PKG_DEPS_INDENT * 2)))
	    fputc(PKG_DEPS_INDENT_LINK, pipe);
	 else
	    fputc(PKG_DEPS_INDENT_CHAR, pipe);
      }

      fputc((STAILQ_NEXT(entry, entries) != NULL) ? PKG_DEPS_INDENT_LINK 
						  : PKG_DEPS_INDENT_CHAR, pipe);

      tofree = bufp = STRNDUP(entry->pkg_fields[PKG_CAT], INDEX_MAX_LINE_LEN);
      
      if (tofree == NULL) {
	 fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	 ERRX;
      }

      token = strsep(&bufp, " ");

      fprintf(pipe, ds, token, entry->pkg_fields[PKG_NAME]);

      free(token);
      
      STAILQ_FOREACH(deps_entry, &entry->deps_head, entries) {
	 tmp_is_last = (STAILQ_NEXT(deps_entry, entries) == NULL) && \
		       (STAILQ_NEXT(entry, entries) == NULL);
	 pkg_display_deps_full_tree(deps_entry, pipe, 0,                    \
			           (entry->flags & PKG_FLAGS_SELECTED) ||   \
			           (pkg_entry->flags & PKG_FLAGS_SELECTED) || selected, \
				   tmp_is_last);
	 
	 indent -= PKG_DEPS_INDENT;
      }
   }

   indent -= PKG_DEPS_INDENT;

   return(res);
}

/* pkg_find_cats
 *
 * Finds the primary categories from the INDEX file,
 * and the number of packages in each of them.
 *
 * Returns the number of categories.
 */
int
pkg_find_cats(void)
{
   PKG     *entry, *cat_entry;
   FILE    *fp;
   char     bigbuf[INDEX_MAX_LINE_LEN]; /* a buffer big enough to store one line from INDEX */
   char    *tofree, *token, *bufp,
	   *tofree_cat, *token_cat, *bufp_cat;
   int      i, flag;
   unsigned cats_num;

   entry = (PKG *) malloc(sizeof(PKG));

   if (entry == NULL) {
      fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
      ERRX;
   }

   entry->flags = 0;
   entry->pkg_num = 0;
   STAILQ_INIT(&entry->deps_head);

   for (i = 0; i < PKG_FIELDS; i++)
      entry->pkg_fields[i] = NULL;

   /* fetch the INDEX file if it is missing */
   if (!file_exists(config_get_option(CONFIG_INDEX_FILE))) 
      fetch_index_file();
            
   if ((fp = fopen(config_get_option(CONFIG_INDEX_FILE), "r")) == NULL) {
      fprintf(stderr, "%s: Cannot open INDEX file.\n", PROGNAME);
      ERRX;
   }
   
   cats_num = 0;

   /* make sure we'll read one whole line from INDEX */
   while (!feof(fp) && fgets(bigbuf, INDEX_MAX_LINE_LEN, fp)) {
      /* remove CR/LF from the end of the string */
      bigbuf[STRNLEN(bigbuf, INDEX_MAX_LINE_LEN) - 1] = '\0';

      tofree = bufp = STRNDUP(bigbuf, INDEX_MAX_LINE_LEN);

      if (tofree == NULL) {
	 fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
         ERRX;
      }

      /*
       * get the fields of the package from the line we just read
       * each package field in INDEX is separated by PKG_FIELD_SEPARATOR char
       *
       * we read and through away all elements, except for the pkg_fields[PKG_CAT] one
       */
      for (i = 0; i <= PKG_CAT; i++)
	 token = strsep(&bufp, PKG_FIELDS_SEPARATOR);

      tofree_cat = bufp_cat = STRNDUP(token, MAXPATHLEN);
      token_cat  = strsep(&bufp_cat, " ");

      entry->pkg_fields[PKG_CAT] = STRNDUP(token_cat, INDEX_MAX_LINE_LEN);

      if (entry->pkg_fields[PKG_CAT] == NULL) {
	 fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	 ERRX;
      }
      
      free(tofree);
      free(tofree_cat);

      flag = 0;

      /* check if the category exists already */
      STAILQ_FOREACH(cat_entry, &pkg_head, entries)
	 if (!strncmp(entry->pkg_fields[PKG_CAT], cat_entry->pkg_fields[PKG_CAT], 
		      INDEX_MAX_LINE_LEN)) {
	    flag = 1;
	    break;
	 }

      if (flag)
	 cat_entry->pkg_num++;
      else {
	 pkg_insert(entry, 0);
	 cats_num++;
      }

      free(entry->pkg_fields[PKG_CAT]);
   }
      
   free(entry);
   fclose(fp);

   return(cats_num);
}

/*
 * pkg_find_deps
 *
 * Creates a one-level tail of packages, plus their dependencies
 *
 * Returns the number of packages selected for install.
 */
int
pkg_find_deps(void)
{
   PKG	 *entry, *tmp_entry, *newentry;
   char	 *tofree, *bufp, *token;
   int	  i, flag, total_pkgs;

   total_pkgs = 0;

   /* 
    * put all packages, plus their dependencies in one-level tail,
    * that we can later use to build the full deps tree
    */
   STAILQ_FOREACH(entry, &pkg_head, entries) { 
      /* check for duplicate packages in the pkg_single_head tail */
      flag = 0;

      STAILQ_FOREACH(tmp_entry, &pkg_single_head, entries)
	 if (!strncmp(tmp_entry->pkg_fields[PKG_NAME], entry->pkg_fields[PKG_NAME], MAXPATHLEN)) {
	    flag = 1;
	    break;
	 }

      if (flag)
	 continue;

      total_pkgs++;

      newentry = (PKG *) malloc(sizeof(PKG));
      
      if (newentry == NULL) {
	 fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	 ERRX;
      }

      newentry->flags = entry->flags;
      newentry->pkg_num = 1;
      STAILQ_INIT(&newentry->deps_head);

      for (i = 0; i < PKG_FIELDS; i++) 
	 newentry->pkg_fields[i] = NULL;

      newentry->pkg_fields[PKG_NAME] = STRNDUP(entry->pkg_fields[PKG_NAME], MAXPATHLEN);
      newentry->pkg_fields[PKG_CAT]  = STRNDUP(entry->pkg_fields[PKG_CAT], MAXPATHLEN);
      
      if ((newentry->pkg_fields[PKG_NAME] == NULL) || (newentry->pkg_fields[PKG_CAT] == NULL)) {
	 fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	 ERRX;
      }

      STAILQ_INSERT_TAIL(&pkg_single_head, newentry, entries);

      if (entry->pkg_fields[PKG_RDEPS] == NULL)
	 continue;

      tofree = bufp = STRNDUP(entry->pkg_fields[PKG_RDEPS], INDEX_MAX_LINE_LEN);

      if (tofree == NULL) {
	 fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	 ERRX;
      }

      /* get the dependencies of 'entry' and save them too in the tail */
      while ((token = strsep(&bufp, " ")) != NULL) {
	 /* check for duplicate packages in the pkg_single_head tail */
	 flag = 0;
      
	 STAILQ_FOREACH(tmp_entry, &pkg_single_head, entries)
	    if (!strncmp(tmp_entry->pkg_fields[PKG_NAME], token, MAXPATHLEN)) {
	       flag = 1;
	       break;
	    }

	 if (flag)
	    continue;

	 total_pkgs++;

	 newentry = (PKG *) malloc(sizeof(PKG));

	 if (newentry == NULL) {
	    fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	    ERRX;
	 }

	 newentry->flags = 0;
	 newentry->pkg_num = 1;
	 STAILQ_INIT(&newentry->deps_head);

	 for (i = 0; i < PKG_FIELDS; i++)
	    newentry->pkg_fields[i] = NULL;

	 newentry->pkg_fields[PKG_NAME] = STRNDUP(token, MAXPATHLEN);

	 if (newentry->pkg_fields[PKG_NAME] == NULL) {
	    fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	    ERRX;
	 }

	 STAILQ_INSERT_TAIL(&pkg_single_head, newentry, entries);
      }

      free(tofree);
   }
   
   return(total_pkgs);
}

/* pkg_remove_extension
 *
 * Removes the PKG_EXTENSION part from a package name if any
 *
 * Arguments:
 *    - pkg_entry  - the package entry we want to remove
 *		     the extension from.
 *
 */
void
pkg_remove_extension(PKG *pkg_entry)
{
   char *tofree, *bufp, *token;

   tofree = bufp = STRNDUP(pkg_entry->pkg_fields[PKG_NAME], MAXPATHLEN);

   if (tofree == NULL) {
      fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
      ERRX;
   }

   token = strrchr(bufp, '.');

   if (!strncmp(token, PKG_EXTENSION, MAXPATHLEN)) {
      bufp[STRNLEN(bufp, MAXPATHLEN) - STRNLEN(PKG_EXTENSION, MAXPATHLEN)] = '\0';

      free(pkg_entry->pkg_fields[PKG_NAME]);
      pkg_entry->pkg_fields[PKG_NAME] = STRNDUP(bufp, MAXPATHLEN);
      
      if (pkg_entry->pkg_fields[PKG_NAME] == NULL) {
	 fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	 ERRX;
      }
   }
   
   free(tofree);
}

/* is_package_installed
 *
 * Iterates through the package tails in order
 * to find out if a package was already installed or not.
 *
 * If a package is installed already this function also
 * sets the PKG_FLAGS_INSTALLED flag.
 *
 * Arguments:
 *
 *    - pkg_entry	- the package that will be queried if it's installed
 *    - set_flag	- when true sets the PKG_FLAG_INSTALLED flag
 *    - total_pkgs	- self-explaining
 *
 * Returns true if `pkg_entry' is installed, false otherwise.
 */
int
is_pkg_installed(PKG *pkg_entry, int set_flag, int total_pkgs)
{
   char	  buf[MAXPATHLEN];
   char	 *pkg_info, *pkg_info_args;
   int	  result;
   static int current_pkgs = 0;

   result        = 0;
   pkg_info      = config_get_option(CONFIG_PKG_INFO_CMD);
   pkg_info_args = (config_get_option(CONFIG_PKG_INFO_CMD_ARGS) == NULL) ? \
		   " " : config_get_option(CONFIG_PKG_INFO_CMD_ARGS);

   snprintf(buf, sizeof(buf), "%s %s %s", \
	    pkg_info, pkg_info_args, pkg_entry->pkg_fields[PKG_NAME]);

   result = system(buf);

   if (set_flag && !result)
      pkg_entry->flags |= PKG_FLAGS_INSTALLED;
   
   /* display a progress if total_pkgs > 0 
    * total_pkgs > 0 during the initial package search that is done before
    * installation of any package.
    *
    * After the installation of the packages, is_pkg_installed() is 
    * called again to check whether or not there were any newly installed
    * packages when the dependency resolving is not under pkg_add_it(1) control
    */
   current_pkgs++;

   if (total_pkgs) 
      printf("==>> Querying database for installed packages: %.2f%% complete.\r", \
	       ((float)current_pkgs / (float)total_pkgs) * 100);

   return(!result);
}

/* pkg_install_packages
 *
 * Installs all packages that were selected for install.
 *
 * Arguments:
 *
 *    - install_options - contains the different install options,
 *			  like whether it is remote or local install,
 *			  use a pager or not, etc...
 */
void
pkg_install_packages(int install_options)
{
   PKG   *entry;
   struct pkg_head_type *headp;
   FILE  *fp, *pipe;
   char   buf[MAXPATHLEN], line[MAXPATHLEN];
   char	 *pkg_url, *pkg_install_cmd, *pkg_install_cmd_args, 
	 *pager, *token, *tofree, *bufp; 
   int    choice, deps_full_tree,
	  abort, display_files;

   pager   = token = tofree = bufp = NULL;
   pkg_url = pkg_install_cmd = pkg_install_cmd_args = NULL;

   /* deps_full_tree can be true, only when doing a remote install */
   deps_full_tree = (install_options & OPTION_LOCAL_INSTALL) ? 0 : \
		     config_get_bool_option(CONFIG_DEPS_FULL_TREE);

   /* do nothing if there are no packages for install */
   if (pkg_show_deps(install_options) == 0)
      return;

   /* skip installation process if --show-deps option is specified */
   if (install_options & OPTION_SHOW_DEPS)
      return;

   choice = yesno_prompt("==>> Proceed with installation [y/n]: ");

   if (choice == 'n' || choice == 'N')
      return;

   headp = (deps_full_tree) ? (struct pkg_head_type *) &pkg_head
			    : (struct pkg_head_type *) &pkg_single_head;

   if (install_options & OPTION_REMOTE_INSTALL) {
      pkg_install_cmd      = config_get_option(CONFIG_PKG_INSTALL_CMD_REMOTE);
      pkg_install_cmd_args = (config_get_option(CONFIG_PKG_INSTALL_CMD_REMOTE_ARGS) == NULL) ? \
				 " " : config_get_option(CONFIG_PKG_INSTALL_CMD_REMOTE_ARGS);
      pkg_url		   = config_get_option(CONFIG_PKG_URL);
   } else {
	 pkg_install_cmd      = config_get_option(CONFIG_PKG_INSTALL_CMD);
	 pkg_install_cmd_args = (config_get_option(CONFIG_PKG_INSTALL_CMD_ARGS) == NULL) ? \
				 " " : config_get_option(CONFIG_PKG_INSTALL_CMD_ARGS);
   }

   STAILQ_FOREACH(entry, headp, entries) {
      abort = 0;

      /* 
       * skip packages that are already selected, installed or packages
       * that are not root packages (packages that were selected during initial search)
       */
      if ((entry->flags & PKG_FLAGS_SELECTED) || (entry->flags & PKG_FLAGS_INSTALLED) || \
	    !(entry->flags & PKG_FLAGS_ROOT_PKG))
	 continue;

      /* install package dependencies if any and if deps_full_tree=true */
      if ((deps_full_tree == 1) && (entry->pkg_fields[PKG_RDEPS] != NULL)) {
	 abort = pkg_install_deps(entry, 1);

	 if (abort == 'y' || abort == 'Y')
	    printf("\n==>> Skipping package installation for %s due to failed dependencies\n",
		     entry->pkg_fields[PKG_NAME]);

	 /* in all cases, we should continue to the next package install */
	 continue;
      }

      tofree = bufp = STRNDUP(entry->pkg_fields[PKG_CAT], MAXPATHLEN);

      if (tofree == NULL) {
	 fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	 ERRX;
      }

      token = strsep(&bufp, " ");

      if (install_options & OPTION_LOCAL_INSTALL) 
	 snprintf(buf, sizeof(buf), "%s %s %s/%s%s",
			pkg_install_cmd,
			pkg_install_cmd_args,
			entry->pkg_fields[PKG_CAT],
			entry->pkg_fields[PKG_NAME],
			PKG_EXTENSION);
      else
	 snprintf(buf, sizeof(buf), "%s %s %s/%s/%s%s", 
			pkg_install_cmd,
			pkg_install_cmd_args,
			pkg_url,
			token,
			entry->pkg_fields[PKG_NAME],
			PKG_EXTENSION);

      printf("\n==>> Installing package %s ...\n", entry->pkg_fields[PKG_NAME]);

      (system(buf) == 0) ? entry->flags |= PKG_FLAGS_INSTALL_RESULT : 0;

      free(tofree);
   }

   /*
    * Perform some post-install actions like 
    * displaying the post-install messages and summary of installation.
    *
    * Here we only use pkg_single_head, since it contains all the packages
    * that were about to be installed, so no need to go in recursion.
    *
    * In order to find the installed packages we compare the package state
    * before and after package installation - if it change from not installed
    * to installed, then we have a package to process.
    */
      
   display_files = 0;

   STAILQ_FOREACH(entry, &pkg_single_head, entries) {
      /* check if the package state has change from not installed to installed */
      if (!(entry->flags & PKG_FLAGS_INSTALLED) && is_pkg_installed(entry, 0, 0)) {
	 /* check if the package has a +DISPLAY file */
	 snprintf(buf, sizeof(buf), "%s/%s/%s",
			   PKG_DB_PATH,
			   entry->pkg_fields[PKG_NAME],
			   PKG_DISPLAY_FILE);

	 display_files += file_exists(buf);
      }
   }

   if (display_files) {
      pager = getenv("PAGER");

      pipe = (pager == NULL) ? stdout : popen(pager, "w");
   
      if (pipe == NULL)
	 pipe = stdout;
      
      fprintf(pipe, "==>> Displaying post-install messages...\n");
	 
      STAILQ_FOREACH(entry, &pkg_single_head, entries) {
	 if (!(entry->flags & PKG_FLAGS_INSTALLED) && is_pkg_installed(entry, 0, 0)) {  
	    /* show the +DISPLAY file */
	    snprintf(buf, sizeof(buf), "%s/%s/%s",
			   PKG_DB_PATH,
			   entry->pkg_fields[PKG_NAME],
			   PKG_DISPLAY_FILE);

	    if (file_exists(buf)) {
	       if ((fp = fopen(buf, "r")) == NULL) {
		  fprintf(stderr, "%s: Cannot open +DISPLAY file in %s.\n", PROGNAME, buf);
		  ERRX;
	       }

	       fprintf(pipe, "\n==>> Displaying post-install message for %s\n\n",
			      entry->pkg_fields[PKG_NAME]);

	       while (!feof(fp) && (fgets(line, MAXPATHLEN - 1, fp) != NULL))
		  fprintf(pipe, line);

		  fclose(fp);
	    }
	 }
      }

      if (pipe != stdout) 
	 pclose(pipe);
   }
	    
   printf("\n==>> Summary of package installation:\n\n");

   STAILQ_FOREACH(entry, &pkg_single_head, entries) 
      if (!(entry->flags & PKG_FLAGS_INSTALLED)) {
	 is_pkg_installed(entry, 1, 0);
	 printf("\t* %-30s installation %s.\n", entry->pkg_fields[PKG_NAME],  \
					  (entry->flags & PKG_FLAGS_INSTALLED) ? \
					  "succeeded" : "failed");
      }
}

/* pkg_install_deps
 *
 * Installs the dependencies for a certain package.
 * This function is used only when doing remote installation.
 *
 * Arguments:
 *
 *    - pkg_entry - specifies the package that we need
 *		    to have it's dependencies installed.
 *    - root_pkg  - when true specifieds that this is a root package
 */
int
pkg_install_deps(PKG *pkg_entry, int root_pkg)
{
   PKG	       *entry;
   static char *pkg_url, *pkg_install_cmd, *pkg_install_cmd_args;
   char	        buf[MAXPATHLEN];
   char	       *tofree, *token, *bufp;
   static int   init = 0, abort = 0;

    /* get the install command and it's arguments only once */
   if (!init) {
      init = 1;
      pkg_install_cmd	   = config_get_option(CONFIG_PKG_INSTALL_CMD_REMOTE);
      pkg_install_cmd_args = (config_get_option(CONFIG_PKG_INSTALL_CMD_REMOTE_ARGS) == NULL) ? \
			      " " : config_get_option(CONFIG_PKG_INSTALL_CMD_REMOTE_ARGS);
      pkg_url		   = config_get_option(CONFIG_PKG_URL);
   }

   /* check if some of the previous dependencies install has failed */
   abort = (root_pkg == 1) ? 0 : abort;

   if (abort == 'y' || abort == 'Y')
      return(abort);

   printf("==>> Starting dependencies installation for package %s ...\n", \
		     pkg_entry->pkg_fields[PKG_NAME]);

   if (*(pkg_entry->pkg_fields[PKG_RDEPS]) == '\0')
      printf("==>> Package %s has no registered dependencies ...\n", 
		     pkg_entry->pkg_fields[PKG_NAME]);

   STAILQ_FOREACH(entry, &pkg_entry->deps_head, entries) {
      if ((entry->flags & PKG_FLAGS_SELECTED) || (entry->flags & PKG_FLAGS_INSTALLED))
	 continue;

      pkg_install_deps(entry, 0);
   }
      
   /* check if some of the previous dependencies install has failed */
   if (abort == 'y' || abort == 'Y')
      return(abort);
   
   tofree = bufp = STRNDUP(pkg_entry->pkg_fields[PKG_CAT], MAXPATHLEN);

   if (tofree == NULL) {
      fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
      ERRX;
   }

   token = strsep(&bufp, " ");

   /* install the package */
   snprintf(buf, sizeof(buf), "%s %s %s/%s/%s%s", 
		     pkg_install_cmd,
		     pkg_install_cmd_args,
		     pkg_url,
		     token,
		     pkg_entry->pkg_fields[PKG_NAME],
		     PKG_EXTENSION);

   printf("\n==>> Installing package %s ...\n", pkg_entry->pkg_fields[PKG_NAME]);

   abort = system(buf);
      
   free(tofree);

   (abort == 0) ? pkg_entry->flags |= PKG_FLAGS_INSTALL_RESULT \
	        : (abort = pkg_deps_abort_install(pkg_entry));

   return(abort);
}

/* pkg_deps_abort_install
 *
 * Displays a warning message and asks the user
 * if he/she wants to proceed with the package
 * installation.
 *
 * Returns 'y' or 'n' depending on the selected choice
 */
int
pkg_deps_abort_install(PKG *pkg_entry)
{
   int choice;

   printf("\n==>> WARNING: Dependency installation failed.\n");
   printf("\n==>> The following package failed to install: %s\n", \
	  pkg_entry->pkg_fields[PKG_NAME]);

   printf("\n==>> It is highly recommended that you abort the package dependencies\n");
   printf("     installation, in order to avoid any corruption of the package database.\n");

   printf("\n==>> Continuing installing the rest of the package dependencies\n");
   printf("     may result in a broken package database. If you choose to\n");
   printf("     continue with the installation process, you will have to deal\n");
   printf("     with resolving all package dependencies by yourself later on.\n\n");
    
   choice = yesno_prompt("==>> Abort dependencies installation [y/n]: ");

   return(choice);
} 

/* pkg_finish
 *
 * Deallocates the memory used by the tails.
 */ 
void
pkg_finish(void)
{
   PKG *entry, *deps_entry, 
       *entry_safe, *deps_entry_safe;
   int  i;
 
   STAILQ_FOREACH_SAFE(entry, &pkg_head, entries, entry_safe) {
      STAILQ_FOREACH_SAFE(deps_entry, &entry->deps_head, entries, deps_entry_safe)
	 pkg_free_deps(deps_entry, entry);

      for(i = 0; i < PKG_FIELDS; i++)
	 if (entry->pkg_fields[i] != NULL)
	    free(entry->pkg_fields[i]); 

      STAILQ_REMOVE(&pkg_head, entry, pkg, entries);
      free(entry);
   }

   STAILQ_FOREACH_SAFE(entry, &pkg_single_head, entries, entry_safe) {
      for(i = 0; i < PKG_FIELDS; i++)
	 if (entry->pkg_fields[i] != NULL)
	    free(entry->pkg_fields[i]); 

      STAILQ_REMOVE(&pkg_single_head, entry, pkg, entries);
      free(entry);
   }
}

/* pkg_free_deps
 *
 * Deallocates the memory used by all dependencies of a certain package.
 *
 * Arguments:
 *
 *    - deps_entry - dependency package to remove
 *    - pkg_entry  - the package 'deps_entry' is a dependency of.
 */
void
pkg_free_deps(PKG *deps_entry, PKG *pkg_entry)
{
   PKG *entry, *entry_safe;
   int  i;

   /* check if the package has been selected/copied already */
   STAILQ_FOREACH_SAFE(entry, &deps_entry->deps_head, entries, entry_safe)
      pkg_free_deps(entry, deps_entry);

   for (i = 0; i < PKG_FIELDS; i++)
      if (deps_entry->pkg_fields[i] != NULL)
	 free(deps_entry->pkg_fields[i]); 

   STAILQ_REMOVE(&pkg_entry->deps_head, deps_entry, pkg, entries);
   free(deps_entry);
}

/* pkg_local_install
 *
 * performs a local package installation and does some checks
 * of the provided command-line arguments
 *
 * Returns the number of found packages.
 */
int
pkg_local_install(int argc, char *argv[], int optind, int options)
{
   char *cwd;
   int  numfiles;

   if (argv[optind] == NULL) {
       fprintf(stderr, "==>> You have to specify <pkg-name>\n");
       fprintf(stderr, "==>> Run %s --help for a list of available options\n", PROGNAME);
       exit(1);
   }

   if (argc > (optind +1)) {
      fprintf(stderr, "==>> Too many arguments passed\n");
      fprintf(stderr, "==>> Run %s --help for a list of available options\n", PROGNAME);
      exit(1);
   }

   if (!dir_exists(argv[optind - 1])) {
      fprintf(stderr, "==>> '%s' is not a directory\n", argv[optind]);
      fprintf(stderr, "==>> Run %s --help for a list of available options\n", PROGNAME);
      exit(1);
   }
            
   if ((cwd = getcwd(NULL, MAXPATHLEN + 1)) == NULL) {
      fprintf(stderr, "%s: Cannot get the current working directory.\n", PROGNAME);
      WARNX;
   }

   numfiles = scan_directory(argv[optind - 1], argv[optind], options);

   chdir(cwd);
   free(cwd);

   return(numfiles);
}

/* pkg_remote_install
 *
 * perfoms an INDEX scan and does some checks
 * of the provided command-line arguments
 *
 * Returns the number of found packages.
 */
int
pkg_remote_install(int argc, char *argv[], int optind, int options)
{
   int numfiles;

   if (argv[optind - 1] == NULL) {
      fprintf(stderr, "==>> You have to specify <pkg-name>\n");
      fprintf(stderr, "==>> Run %s --help for a list of available options\n", PROGNAME);
      exit(1);
   }

   if (argc > (optind + 1)) {
      fprintf(stderr, "==>> Too many arguments passed\n");
      fprintf(stderr, "==>> Run %s --help for a list of available options\n", PROGNAME);
      exit(1);
   }
            
   numfiles = scan_index_file(argv[optind - 1], ((argv[optind] == NULL) ? NULL : argv[optind]), options);

   return(numfiles);
}

/* pkg_show_deps
 *
 * Displays package dependencies.
 *
 * This is the main function that is used to find all package deps
 *
 * Returns > 0 if there are packages to be installed, false otherwise.
 */
int
pkg_show_deps(int install_options)
{
   PKG   *entry, *tmp_entry;
   FILE  *pipe;
   char	 *pager;
   int    deps_full_tree,
	  total_pkgs, increment_pkgs, to_be_installed;

   total_pkgs = increment_pkgs = 0;

   deps_full_tree = (install_options & OPTION_LOCAL_INSTALL) ? 0 : \
		    config_get_bool_option(CONFIG_DEPS_FULL_TREE);

   /* save all packages, plus their dependencies in one-level tail */
   total_pkgs = pkg_find_deps();

   printf("\n");

   /* remove package extension when doing local install */
   if (install_options & OPTION_LOCAL_INSTALL)
      STAILQ_FOREACH(entry, &pkg_single_head, entries)
	 pkg_remove_extension(entry);

   STAILQ_FOREACH(entry, &pkg_single_head, entries)
      is_pkg_installed(entry, 1, total_pkgs);

   /* 
    * expand the pkg_head tail, so that it contains all
    * package dependencies if doing a remote install
    */
   if (deps_full_tree) {
      pkg_get_deps_info_from_index(total_pkgs);

      STAILQ_FOREACH(entry, &pkg_head, entries) {
	 if (entry->pkg_fields[PKG_RDEPS] != NULL)
	    pkg_build_deps_full_tree(entry);

	 /* copy the flags for the root package */
	 STAILQ_FOREACH(tmp_entry, &pkg_single_head, entries)
	    if (!strncmp(tmp_entry->pkg_fields[PKG_NAME], entry->pkg_fields[PKG_NAME], \
			MAXPATHLEN)) {
	       entry->flags = tmp_entry->flags;
	       break;
	    }
      }
   }

   /* display the packages and their dependencies, if any */
   pager = getenv("PAGER");

   pipe = ((pager == NULL) || !(install_options & OPTION_PAGER)) ? stdout : popen(pager, "w");

   if (pipe == NULL)
      pipe = stdout;

   fprintf(pipe, "\n\n==>> The following packages have been selected for install:\n\n");

   to_be_installed = 0;

   if (deps_full_tree) 
      STAILQ_FOREACH(entry, &pkg_head, entries) {
	 to_be_installed += pkg_display_deps_full_tree(entry, pipe, 1, (entry->flags & PKG_FLAGS_SELECTED), 1);
	 fprintf(pipe, "\n");
      }
   else
      to_be_installed = pkg_display_deps(pipe);

   fprintf(pipe, "\n");
      
   if (install_options & OPTION_PAGER)
      pclose(pipe);

   if (!to_be_installed) {
      printf("==>> Nothing to install. Everything seems to be installed already.\n\n");
      return(0);
   }

   if (install_options & OPTION_PAGER)
      printf("\n");
   
   printf("==>> Number of packages to be installed: %d\n\n", to_be_installed);

   return(1);
}

/* pkg_build_deps_full_tree
 *
 * Builds a tree of package dependencies in multi-level depth tails
 *
 * This function uses pkg_single_head in order to build the full 
 * dependency tree of packages. 
 */
void
pkg_build_deps_full_tree(PKG *pkg_entry)
{
   PKG  *tmp_entry, *newentry;
   char *tofree, *bufp, *token;
   int   i;

   tofree = bufp = STRNDUP(pkg_entry->pkg_fields[PKG_RDEPS], INDEX_MAX_LINE_LEN);

   if (tofree == NULL) {
      fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
      ERRX;
   }

   while ((token = strsep(&bufp, " ")) != NULL) {
      /* find the package dependency from the 'pkg_single_head' tail 
       * 
       * if we have a match, copy the package, and set the PKG_FLAGS_COPIED flag
       */
      STAILQ_FOREACH(tmp_entry, &pkg_single_head, entries) {
	 if (!strncmp(token, tmp_entry->pkg_fields[PKG_NAME], MAXPATHLEN)) {
	    newentry = (PKG *) malloc(sizeof(PKG));

	    if (newentry == NULL) {
	       fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	       ERRX;
	    }

	    newentry->flags = tmp_entry->flags;
	    newentry->pkg_num = 1;
	    STAILQ_INIT(&newentry->deps_head);

	    for (i = 0; i < PKG_FIELDS; i++) {
	       if (tmp_entry->pkg_fields[i] != NULL) {
		  newentry->pkg_fields[i] = STRNDUP(tmp_entry->pkg_fields[i], INDEX_MAX_LINE_LEN);

		  if (newentry->pkg_fields[i] == NULL) {
		     fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
		     ERRX;
		  }
	       } else
		  newentry->pkg_fields[i] = NULL;
	    }

	    /* make the package selected, if it has been copied already */
	    if (tmp_entry->flags & PKG_FLAGS_COPIED)
	       newentry->flags  |= PKG_FLAGS_SELECTED;
	    else
	       tmp_entry->flags |= PKG_FLAGS_COPIED;

	    STAILQ_INSERT_TAIL(&pkg_entry->deps_head, newentry, entries);

	    if (newentry->pkg_fields[PKG_RDEPS] != NULL)
	       pkg_build_deps_full_tree(newentry);
	 }
      }
   }

   free(tofree);
}

/* pkg_get_deps_info_from_index
 *
 * Finds the information for packages selected for install from the INDEX file
 */
void
pkg_get_deps_info_from_index(int total_pkgs)
{
   PKG	       *pkg_entry, *index_entry;
   static char  bigbuf[INDEX_MAX_LINE_LEN]; 
   FILE	       *fp;
   char	       *tofree, *token, *bufp;
   int	        i, pkg_field_index, current_pkgs;

   current_pkgs = 0;
   index_entry = (PKG *) malloc(sizeof(PKG));

   if (index_entry == NULL) {
      fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
      ERRX;
   }

   index_entry->flags = 0;
   index_entry->pkg_num = 0;
   STAILQ_INIT(&index_entry->deps_head);

   for (i = 0; i < PKG_FIELDS; i++)
      index_entry->pkg_fields[i] = NULL;

   /* fetch the INDEX file if it is missing */
   if (!file_exists(config_get_option(CONFIG_INDEX_FILE))) 
      fetch_index_file();
            
   if ((fp = fopen(config_get_option(CONFIG_INDEX_FILE), "r")) == NULL) {
      fprintf(stderr, "%s: Cannot open INDEX file.\n", PROGNAME);
      ERRX;
   }
   
   printf("\n");

   /* make sure we'll read one whole line from INDEX */
   while (!feof(fp) && fgets(bigbuf, INDEX_MAX_LINE_LEN, fp) && (current_pkgs != total_pkgs)) {
      printf("==>> Gathering package(s) information from INDEX: %.2f%% complete.\r",
		     (((float)current_pkgs + 1) / (float)total_pkgs) * 100); 

      pkg_field_index = 0;

      bigbuf[STRNLEN(bigbuf, INDEX_MAX_LINE_LEN) - 1] = '\0';

      tofree = bufp = STRNDUP(bigbuf, INDEX_MAX_LINE_LEN);

      if (tofree == NULL) {
	 fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
         ERRX;
      }

      /*
       * get the fields from the line we just read.
       * each package field in INDEX is separated by
       * the PKG_FIELD_SEPARATOR character.
       */
      while ((token = strsep(&bufp, PKG_FIELDS_SEPARATOR)) != NULL) {
         index_entry->pkg_fields[pkg_field_index++] = STRNDUP(token, INDEX_MAX_LINE_LEN);
         
         if (index_entry->pkg_fields[pkg_field_index - 1] == NULL) {
	    fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	    ERRX;
	 }
      }

      free(tofree);
   
      /* check if index_entry is one of our packages, contained in the pkg_single_head tail,
       * if we find a match, we save the PKG_CAT and PKG_RDEPS elements. 
      */
      STAILQ_FOREACH(pkg_entry, &pkg_single_head, entries) {
	 if (!strncmp(pkg_entry->pkg_fields[PKG_NAME], index_entry->pkg_fields[PKG_NAME], MAXPATHLEN)) {
	    current_pkgs++;

	    /* 
	     * category is already present for the root packages,
	     * the ones that have the PKG_FLAGS_ROOT_PKG flag set
	     */
	    if (!(pkg_entry->flags & PKG_FLAGS_ROOT_PKG)) {
	       /* get the primary category */
	       tofree = bufp = STRNDUP(index_entry->pkg_fields[PKG_CAT], MAXPATHLEN);

	       if (tofree == NULL) {
		  fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
		  ERRX;
	       }

	       token = strsep(&bufp, " ");
	    
	       pkg_entry->pkg_fields[PKG_CAT] = STRNDUP(token, MAXPATHLEN);
	    
	       if (pkg_entry->pkg_fields[PKG_CAT] == NULL) {
		  fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
		  ERRX;
	       }

	       free(tofree);
	    }

	    pkg_entry->pkg_fields[PKG_RDEPS] = STRNDUP(index_entry->pkg_fields[PKG_RDEPS], INDEX_MAX_LINE_LEN);
	    
	    if (pkg_entry->pkg_fields[PKG_RDEPS] == NULL) {
	       fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	       ERRX;
	    }

	    break;
	 }
      }

      /* free the index_entry elements */
      for (i = 0; i < PKG_FIELDS; i++)
	 if (index_entry->pkg_fields[i] != NULL)
	    free(index_entry->pkg_fields[i]);
   }

   free(index_entry);
}

