/*-
 * Copyright (c) 2010  Marin Atanasov Nikolov  <dnaeon@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/param.h>

#include <stdio.h>
#include <regex.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "main.h"
#include "config.h"
#include "pkg.h"
#include "scan_index.h"

/* file_exists
 *
 * checks for existence of a specified file
 */
int
file_exists(const char *filename)
{
   return(access(filename, F_OK) ? 0 : 1);
}

/* fetch_index_file
 *
 * fetches the INDEX file if it doesn't exist in the the location
 * specified by CONFIG_PKG_INDEX_FILE in the configuration file.
 */
void
fetch_index_file(void)
{
   char indexpath[MAXPATHLEN + 1];
         
   fprintf(stdout, "==>> INDEX file was not found in %s, I'll fetch it now...\n\n",
	    config_get_option(CONFIG_INDEX_FILE));
   
   snprintf(indexpath, sizeof(indexpath), "%s %s", 
	    config_get_option(CONFIG_FETCH_CMD), config_get_option(CONFIG_FETCH_CMD_ARGS));
  
   system(indexpath);
}

/* scan_index_file
 *
 * scans the INDEX file for a package pattern
 *
 * Returns the number of found packages
 */
int
scan_index_file(const char *pattern, const char *cat, int options)
{
   PKG	   *entry;
   regex_t  myregex;
   FILE	   *fp;
   char	    bigbuf[INDEX_MAX_LINE_LEN]; /* a buffer big enough to store one line from INDEX */
   char	   *tofree, *token, *bufp;
   unsigned pattern_count;
   int	    pkg_field_index, found_pkg, i;
   
   entry = (PKG *) malloc(sizeof(PKG));

   if (entry == NULL) {
      fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
      ERRX;
   }

   entry->flags = 0;
   entry->pkg_num = 0;
   STAILQ_INIT(&entry->deps_head);

   for (i = 0; i < PKG_FIELDS; i++)
      entry->pkg_fields[i] = NULL;

   if (options & OPTION_REGEX)
      if (regcomp(&myregex, pattern, REG_ICASE | REG_NOSUB | \
	         ((options & OPTION_REGEX_EXTENDED) ? REG_EXTENDED : REG_BASIC)) != 0) {
	 fprintf(stderr, "%s: Cannot compile the given regex.\n", PROGNAME);
	 ERRX;
      }

   /* fetch the INDEX file if it is missing */
   if (!file_exists(config_get_option(CONFIG_INDEX_FILE))) 
      fetch_index_file();
            
   if ((fp = fopen(config_get_option(CONFIG_INDEX_FILE), "r")) == NULL) {
      fprintf(stderr, "%s: Cannot open INDEX file.\n", PROGNAME);
      ERRX;
   }
   
   pattern_count = 0;
   
   /* make sure we'll read one whole line from INDEX */
   while (!feof(fp) && fgets(bigbuf, INDEX_MAX_LINE_LEN, fp)) {
      found_pkg = pkg_field_index = 0;

      /* remove CR/LF from the end of the string */
      bigbuf[STRNLEN(bigbuf, INDEX_MAX_LINE_LEN) - 1] = '\0';

      tofree = bufp = STRNDUP(bigbuf, INDEX_MAX_LINE_LEN);

      if (tofree == NULL) {
	 fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	 ERRX;
      }

      /*
       * get the fields of the package from the line we just read
       * each package field in INDEX is separated by PKG_FIELD_SEPARATOR char
       */
      while ((token = strsep(&bufp, PKG_FIELDS_SEPARATOR)) != NULL) {
	 entry->pkg_fields[pkg_field_index++] = STRNDUP(token, INDEX_MAX_LINE_LEN);
	 
	 if (entry->pkg_fields[pkg_field_index - 1] == NULL) {
	    fprintf(stderr, "%s: Cannot allocate memory.\n", PROGNAME);
	    ERRX;
	 }
      }

      free(tofree);
	 
      /* 
       * if there is a matching entry - save it.
       *
       * the search is done by package name and description.
       *
       * if 'cat' != NULL, we search only the packages from category 'cat'
       */
      if (cat != NULL) {
	 /* get the primary category first */
	 tofree = bufp = STRNDUP(entry->pkg_fields[PKG_CAT], INDEX_MAX_LINE_LEN);
	 token  = strsep(&bufp, " ");
	 
	 /* found a matching category, match the package name/desc as well */
	 if(!strncmp(token, cat, INDEX_MAX_LINE_LEN)) {
	    /* use regex? */
	    if (options & OPTION_REGEX) {
	       if ((!regexec(&myregex, entry->pkg_fields[PKG_NAME], 0, NULL, 0)) || \
		   (!regexec(&myregex, entry->pkg_fields[PKG_DESC], 0, NULL, 0)))
		  found_pkg = 1;
	    } else
	       /* perform a simple search */
	       if ((strcasestr(entry->pkg_fields[PKG_NAME], pattern) != NULL) || \
		   (strcasestr(entry->pkg_fields[PKG_DESC], pattern) != NULL))
		  found_pkg = 1;
	 }

	 free(tofree);
      } else
	 /* else, search from any category */
	 /* use regex? */
	 if (options & OPTION_REGEX) {
	    if ((!regexec(&myregex, entry->pkg_fields[PKG_NAME], 0, NULL, 0)) || \
	        (!regexec(&myregex, entry->pkg_fields[PKG_DESC], 0, NULL, 0)))
	       found_pkg = 1;
	 } else
	    /* perform a simple search */
	    if ((strcasestr(entry->pkg_fields[PKG_NAME], pattern) != NULL) || \
	        (strcasestr(entry->pkg_fields[PKG_DESC], pattern) != NULL))
	       found_pkg = 1;
                 
      /* save the matched package */
      if (found_pkg) {
	 pkg_insert(entry, 1);
	 pattern_count++;
      }

      /* free the memory allocated by entry->pkg_fields[] */
      for (i = 0; i < PKG_FIELDS; i++)
	 free(entry->pkg_fields[i]);
   }
      
   free(entry);
   fclose(fp);
   
   if (options & OPTION_REGEX)
      regfree(&myregex);
   
   return(pattern_count);
}

